_G["LootSquid"] = {
    session = {
        loot = {},
        currentLootSlot = 0,
        playersNeeding = {},
        playersGreeding = {}
    },
    debug = { alwaysParseLoot = false },
    settings = { lootThreshold = 0 },
    frames = { mainFrame = nil },
    eventHandlers = {}
}

StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"] = {
    text = "Transfer x to y?",
    button1 = "Yes",
    button2 = "No",
    OnAccept = function(popup)
    -- find candidate for current loot slot
        local shortPlayerName = Ambiguate(popup.playerName, "all")
        for k in pairs(_G.LootSquid.session.loot) do
            if (_G.LootSquid.session.loot[k].candidates
                    and _G.LootSquid.session.loot[k].candidates[shortPlayerName]) then

                local ci = _G.LootSquid.session.loot[k].candidates[shortPlayerName].candidateIndex

                if ci > 0 then
                    LootSquid:TransferLoot(_G.LootSquid.session.currentLootSlot, ci)
                    LootSquid:ClearNeedAndGreeRequests()
                end
            end
        end

    --
    end,
    timeout = 30,
    whileDead = false,
    hideOnEscape = true,
    preferredIndex = 3, -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

function LootSquid:OnLoad(mainFrame)
    mainFrame:ClearAllPoints()
    mainFrame:SetPoint("LEFT", 0)
    mainFrame:SetPoint("TOP", 0)
    self.frames.mainFrame = mainFrame
    self:Initialize()
end

function LootSquid:Print(message)
    DEFAULT_CHAT_FRAME:AddMessage(format("%s: %s", "LootSquid", message))
end

function LootSquid:strStartsWith(String, Start)
    return string.sub(String, 1, string.len(Start)) == Start
end

function LootSquid:strEndsWith(String, End)
    return End == '' or string.sub(String, -string.len(End)) == End
end

-- Sort a Table[] in Lua
-- http://stackoverflow.com/a/15706820
function LootSquid:spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys + 1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys
    if order then
        table.sort(keys, function(a, b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end


function LootSquid:Initialize()

    self:RegisterEvents()

    --    precalculate api rating
    if _G["LootSquidData"] then
        if (_G["LootSquidData"].standings) then
            for k in pairs(_G["LootSquidData"].standings) do
                local standing = _G["LootSquidData"].standings[k].standing
                local itemsReceived = _G["LootSquidData"].standings[k].itemsReceived
                local apiRating = _G["LootSquidData"].standings[k].apiRating
                if (standing and itemsReceived) then
                    if (itemsReceived > 0) then
                        apiRating = standing / itemsReceived
                    else
                        apiRating = standing / 0.5
                    end
                    _G["LootSquidData"].standings[k].apiRating = apiRating
                end
            end
        else
            LootSquid:Print("Error: No standings available in LootSquidData.")
        end
    end
end

function LootSquid:TimerFeedback()
    print("5 seconds passed")
end

function LootSquid:LootIconClick(lootIndex)
    --LootSquid:Print("Selecting lootIndex: " .. tostring(lootIndex))
end

function LootSquid:LootIconEnter(itemLink)
    GameTooltip:SetOwner(_G["LootSquidFrames"].requestFrame.frame, "ANCHOR_CURSOR")
    GameTooltip:SetHyperlink(itemLink)
    GameTooltip:Show()
end

function LootSquid:LootIconLeave()
    GameTooltip:Hide()
end

function LootSquid:TransferLoot(lootSlot, candidateIndex)
    GiveMasterLoot(lootSlot, candidateIndex)
end

-- http://stackoverflow.com/a/2705804
function LootSquid:tablelength(T)
    local count = 0
    for _ in pairs(T) do count = count + 1 end
    return count
end

function LootSquid:CreateBackdropSolid(edgeSize, insets)

    local edgeFile = "Interface\\ChatFrame\\ChatFrameBackground"

    -- backdrops can't handle edgesize of zero with edgeFile defined
    if edgeSize < 1 then
        edgeFile = nil
    end

    local backdrop = {
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = edgeFile,
        tile = true,
        -- size (width or height) of the square repeating background tiles (in pixels)
        tileSize = 32,
        -- thickness of edge segments and square size of edge corners (in pixels)
        edgeSize = edgeSize,
        -- distance from the edges of the frame to those of the background texture (in pixels)
        insets = insets
    }

    return backdrop
end

function LootSquid:CreateButtonStandard(parent, width, height, text)
    local b = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate")
    b:SetText(text)
    b:SetWidth(width)
    b:SetHeight(height)
    b:ClearAllPoints()
    return b
end

function LootSquid:CreateFrameSolid(parent, width, height, backgroundColor, edgeColor, edgeSize, isMoveAble)
    local f = CreateFrame("Frame", nil, parent)
    f:SetWidth(width)
    f:SetHeight(height)
    f:SetBackdrop(self:CreateBackdropSolid(edgeSize))
    f:SetBackdropColor(backgroundColor.r, backgroundColor.g, backgroundColor.b, backgroundColor.a)
    f:SetBackdropBorderColor(edgeColor.r, edgeColor.g, edgeColor.b, edgeColor.a)

    if isMoveAble then

        f:SetMovable(true)
        f:EnableMouse(true)

        f:SetScript("OnMouseDown",
            function(selfFrame, button)
                if button == "LeftButton" and not selfFrame.isMoving then
                    selfFrame:StartMoving()
                    selfFrame.isMoving = true
                end
            end)

        f:SetScript("OnMouseUp",
            function(selfFrame, button)
                if button == "LeftButton" and selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)

        f:SetScript("OnHide",
            function(selfFrame, button)
                if selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)
    end

    return f
end

function LootSquid:OpenLootGui()

    local f = self.frames.window
    if self.frames.window == nil then

        -- create new window

        -- settings
        local headerHeight = 31
        local footerHeight = 20
        local padding = 3

        f = self:CreateFrameSolid(
            UIParent,
            GetScreenWidth() * 0.6,
            GetScreenHeight() * 0.6,
            { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            padding,
            true)

        f:SetPoint("CENTER", 0, 0)
        f:SetFrameStrata("FULLSCREEN")


        local header = self:CreateFrameSolid(
            f,
            f:GetWidth()-padding*2,
            headerHeight,
            { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        header:SetPoint("TOP", 0, -padding)
        header:SetPoint("LEFT", padding, 0)

        local b = self:CreateButtonStandard(header, 25, 25, "x")
        b:SetPoint("RIGHT", -padding, 0)
        b:SetPoint("TOP", 0, -padding)
        b:SetScript("OnClick", function(selfButton, button) f:Hide() end)

        local title = header:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        title:SetText("Loot Squid")
        title:SetPoint("LEFT", 8, 0)

        local footer = self:CreateFrameSolid(
            f,
            f:GetWidth()-6,
            footerHeight,
            { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        footer:SetPoint("BOTTOM", 0, padding)
        footer:SetPoint("LEFT", padding, 0)

        local bodyFrame = self:CreateFrameSolid(
            f,
            (f:GetWidth()-padding*2),
            (f:GetHeight()-padding*4-headerHeight-footerHeight),
            { r = 0.4, g = 0.1, b = 0.1, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        bodyFrame:SetPoint("TOP", 0, -padding-headerHeight-padding)
        bodyFrame:SetPoint("LEFT", padding, 0)

        local lootFrame = self:CreateFrameSolid(
            bodyFrame,
            (bodyFrame:GetWidth()-padding*2) / 2,
            (bodyFrame:GetHeight()-padding*2-headerHeight) / 3,
            { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        lootFrame:SetPoint("TOP", 0, -padding-headerHeight-padding)
        lootFrame:SetPoint("LEFT", padding, 0)

        local overviewFrame = self:CreateFrameSolid(
            bodyFrame,
            (bodyFrame:GetWidth()-padding*2) / 2,
            (bodyFrame:GetHeight()-padding*2-headerHeight) / 3,
            { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        overviewFrame:SetPoint("TOP", 0, -padding-headerHeight-padding)
        overviewFrame:SetPoint("RIGHT", padding, 0)

        self.frames.window = f
    end


    f:Show()

    --    f:SetScript("OnClose",
    --        function(widget)
    --            _G.LootSquid.session.currentLootSlot = 0
    --            --AceGUI:Release(widget)
    --        end)

    --f:SetTitle("Loot Squid v0.1 Alpha - by Viggo (Darksun - Tarren Mill EU) - test")
    --f:SetLayout("Flow")



    --    local topContainer = AceGUI:Create("SimpleGroup")
    --    topContainer:SetFullWidth(true)
    --    topContainer:SetHeight(200) -- probably?
    --    topContainer:SetLayout("Flow")
    --    f:AddChild(topContainer)
    --
    --    local scroll = AceGUI:Create("ScrollFrame")
    --    scroll:SetRelativeWidth(0.5)
    --    scroll:SetHeight(200)
    --    scroll:SetLayout("Flow") -- probably?
    --
    --    _G["LootSquidFrames"].lootFrame = scroll
    --    topContainer:AddChild(scroll)
    --
    --    self:UpdateLootFrame()
    --
    --
    --    local scrollcontainer = AceGUI:Create("SimpleGroup") -- "InlineGroup" is also good
    --    scrollcontainer:SetFullWidth(true)
    --    scrollcontainer:SetFullHeight(true) -- probably?
    --    scrollcontainer:SetLayout("Fill") -- important!
    --    f:AddChild(scrollcontainer)
    --
    --
    --
    --
    --    _G["LootSquidFrames"].requestFrame = scrollcontainer
    --    self:UpdateRequestFrame()
end

function LootSquid:OnUpdateLootTable()
    self:UpdateLootFrame()
end

function LootSquid:UpdateLootFrame()
    if (not _G["LootSquidFrames"].lootFrame) then
        return
    end

    local AceGUI = LibStub("AceGUI-3.0")


    _G["LootSquidFrames"].lootFrame:ReleaseChildren()

    for k in pairs(_G["LootSquidSession"].loot) do

        local v = _G["LootSquidSession"].loot[k]

        --[[
        -- loot[lootSlot] = {
            itemName = itemName,
            itemQuantity = itemQuantity,
            itemQuality = itemQuality,
            itemLocked = itemLocked,
            itemLootType = itemLootType,
            itemLink = itemLink,
            candidates = {}
        }]]

        local lootGroup = AceGUI:Create("SimpleGroup")
        lootGroup:SetLayout("Flow")

        local lootThreshold = GetLootThreshold()

        if v and v.itemLink and v.itemQuality >= lootThreshold then
            local icon = AceGUI:Create("Icon")
            icon:SetCallback("OnClick",
                function(widget)
                --LootSquid:LootIconClick(k)
                end)
            icon:SetCallback("OnEnter",
                function(widget)
                    LootSquid:LootIconEnter(v.itemLink)
                end)
            icon:SetCallback("OnLeave",
                function(widget)
                    LootSquid:LootIconLeave(v.itemLink)
                end)


            icon:SetLabel(v.itemName)
            icon:SetImageSize(40, 40)

            local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice =
            GetItemInfo(v.itemLink)


            icon:SetImage(texture)
            lootGroup:AddChild(icon)

            local button = AceGUI:Create("Button")
            button:SetText("Offer to raid")
            button:SetCallback("OnClick",
                function(widget)
                    LootSquid:ClearNeedAndGreeRequests()
                    _G["LootSquidSession"].currentLootSlot = k
                    SendChatMessage("LootSquid: --- Loot session starting ---",
                        "RAID")
                    SendChatMessage("LootSquid: do you need or greed " .. v.itemLink .. " ? - if so, write either !need <currentItem> or !greed <currentItem> in chat.",
                        "RAID")
                end)
            lootGroup:AddChild(button)

            button = AceGUI:Create("Button")
            button:SetText("Ninja loot!")
            button:SetCallback("OnClick",
                function(widget)
                end)
            lootGroup:AddChild(button)
        end

        _G["LootSquidFrames"].lootFrame:AddChild(lootGroup)
    end
end

function LootSquid:AddStandingsHeaderRow(frame)

    local AceGUI = LibStub("AceGUI-3.0")

    local row = AceGUI:Create("SimpleGroup")
    row:SetFullWidth(true)
    row:SetLayout("Flow")
    local height = AceGUI:Create("Label")
    height:SetHeight(150)
    height:SetRelativeWidth(0.01)
    row:AddChild(height)

    local backdropR = {
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = "Interface\\ChatFrame\\ChatFrameBackground",
        tile = true,
        -- size (width or height) of the square repeating background tiles (in pixels)
        tileSize = 32,
        -- thickness of edge segments and square size of edge corners (in pixels)
        edgeSize = 2,
        -- distance from the edges of the frame to those of the background texture (in pixels)
        insets = {
            left = 1,
            right = 1,
            top = 1,
            bottom = -1
        }
    }

    row.frame:SetBackdrop(backdropR)
    row.frame:SetBackdropColor(0.15, 0.15, 0.15, 1)
    row.frame:SetBackdropBorderColor(0, 0, 0, 1)

    local label = AceGUI:Create("Label")
    label:SetText("Player")
    label:SetRelativeWidth(0.2)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Spec")
    label:SetRelativeWidth(0.1)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Role")
    label:SetRelativeWidth(0.1)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Current Item")
    label:SetRelativeWidth(0.1)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Attendance")
    label:SetRelativeWidth(0.09)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Items Recv.")
    label:SetRelativeWidth(0.08)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("API")
    label:SetRelativeWidth(0.1)
    row:AddChild(label)

    label = AceGUI:Create("Label")
    label:SetText("Actions")
    label:SetRelativeWidth(0.2)
    row:AddChild(label)

    frame:AddChild(row)
end

function LootSquid:GetStandingsByPlayerName(playerName)
    if (_G["LootSquidData"].standings and _G["LootSquidData"].standings[playerName]) then
        local standing = _G["LootSquidData"].standings[playerName].standing
        local itemsReceived = _G["LootSquidData"].standings[playerName].itemsReceived
        local apiRating = _G["LootSquidData"].standings[playerName].apiRating
        return standing, itemsReceived, apiRating
    end
    return nil
end

function LootSquid:SortStandingsByApiRating(a, b)
    local aStanding, aItemsReceived, aApiRating = LootSquid:GetStandingsByPlayerName(a)
    local bStanding, bItemsReceived, bApiRating = LootSquid:GetStandingsByPlayerName(b)

    if aApiRating and not bApiRating then
        return true
    end

    if bApiRating and not aApiRating then
        return false
    end

    if not aApiRating and not bApiRating then
        return true
    end

    if aApiRating > bApiRating then
        return true
    end

    return false
end

function LootSquid:AddStandingsDataRows(frame, needingOrGreedingPlayers)

    local AceGUI = LibStub("AceGUI-3.0")

    for k in self:spairs(needingOrGreedingPlayers,
        function(t, a, b) return self:SortStandingsByApiRating(a, b) end) do

        local row = AceGUI:Create("SimpleGroup")
        row:SetFullWidth(true)
        row:SetLayout("Flow")
        local height = AceGUI:Create("Label")
        height:SetHeight(150)
        height:SetRelativeWidth(0.01)
        row:AddChild(height)

        local backdropR = {
            bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
            edgeFile = "Interface\\ChatFrame\\ChatFrameBackground",
            tile = true,
            -- size (width or height) of the square repeating background tiles (in pixels)
            tileSize = 32,
            -- thickness of edge segments and square size of edge corners (in pixels)
            edgeSize = 2,
            -- distance from the edges of the frame to those of the background texture (in pixels)
            insets = {
                left = 1,
                right = 1,
                top = 1,
                bottom = -1
            }
        }

        row.frame:SetBackdrop(backdropR)

        row.frame:SetBackdropColor(0.2, 0.2, 0.2, 1)
        row.frame:SetBackdropBorderColor(0, 0, 0, 1)

        --if i < 4 then

        --   row.frame:SetBackdropBorderColor(0,1,0,1)
        --else
        --   row.frame:SetBackdropBorderColor(1,1,0,1)
        --end

        local standing = 0
        local itemsReceived = 0
        local apiRating = 0

        if (_G["LootSquidData"].standings and _G["LootSquidData"].standings[k]) then
            standing = _G["LootSquidData"].standings[k].standing
            itemsReceived = _G["LootSquidData"].standings[k].itemsReceived
            apiRating = _G["LootSquidData"].standings[k].apiRating
        else
            --LootSquid:Print("No standings for player name: "..tostring(k))
        end

        local label = AceGUI:Create("Label")
        label:SetText(k)
        label:SetRelativeWidth(0.2)
        row:AddChild(label)

        label = AceGUI:Create("Label")
        label:SetText("-")
        label:SetRelativeWidth(0.1)
        row:AddChild(label)

        label = AceGUI:Create("Label")
        label:SetText("-")
        label:SetRelativeWidth(0.1)
        row:AddChild(label)

        label = AceGUI:Create("Icon")
        --label:SetText("Current Item")
        label:SetRelativeWidth(0.1)
        row:AddChild(label)

        label = AceGUI:Create("Label")
        label:SetText(standing)
        label:SetRelativeWidth(0.08)
        row:AddChild(label)

        label = AceGUI:Create("Label")
        label:SetText(itemsReceived)
        label:SetRelativeWidth(0.09)
        row:AddChild(label)

        label = AceGUI:Create("Label")
        label:SetText(apiRating)
        label:SetRelativeWidth(0.1)
        row:AddChild(label)

        local actions = AceGUI:Create("SimpleGroup")
        actions:SetRelativeWidth(0.2)
        actions:SetLayout("Flow")
        actions.frame:SetBackdrop(nil)


        local button = AceGUI:Create("Button")
        button:SetFullWidth(true)
        button:SetHeight(20)
        button:SetText("Award")
        button:SetCallback("OnClick",
            function()
                local popup = StaticPopup_Show("LOOTSQUID_AWARD_LOOT", k)
                if popup then
                    popup.playerName = k
                end
            end)
        actions:AddChild(button)

        button = AceGUI:Create("Button")
        button:SetFullWidth(true)
        button:SetHeight(20)
        button:SetText("Whisper for roll")
        button:SetCallback("OnClick", function()
            SendChatMessage("LootSquid: The loot master has requested that you please perform a random roll as part of the loot distribution.",
                "WHISPER",
                nil,
                k)
        end)

        actions:AddChild(button)
        row:AddChild(actions)




        frame:AddChild(row)
    end
end

function LootSquid:UpdateRequestFrame()

    if (not _G.LootSquid.frames.requestFrame) then
        return
    end

    local AceGUI = LibStub("AceGUI-3.0")


    _G.LootSquid.frames.requestFrame:ReleaseChildren()


    local scroll = AceGUI:Create("ScrollFrame")
    scroll:SetLayout("List") -- probably?
    _G.LootSquid.frames.requestFrame:AddChild(scroll)

    self:AddStandingsHeaderRow(scroll)

    local label = AceGUI:Create("InteractiveLabel")
    label:SetText("Players Needing")
    label:SetColor(0.1, 0.8, 0.1)
    label:SetRelativeWidth(1)
    label:SetImageSize(40, 40)
    label:SetHeight(40)
    scroll:AddChild(label)

    self:AddStandingsDataRows(scroll, _G.LootSquid.session.playersNeeding)



    label = AceGUI:Create("InteractiveLabel")
    label:SetText("Players Greeding")
    label:SetColor(0.8, 0.8, 0.1)
    label:SetRelativeWidth(1)
    label:SetImageSize(40, 40)
    label:SetHeight(40)
    scroll:AddChild(label)

    self:AddStandingsDataRows(scroll, _G.LootSquid.session.playersGreeding)
end

function LootSquid:ClearNeedAndGreeRequests()
    _G["LootSquidSession"].playersNeeding = {}
    _G["LootSquidSession"].playersGreeding = {}
    LootSquid:UpdateRequestFrame()
end